# !/usr/bin/env python3

"""Combine property-based testing with contracts of a Python module."""

import argparse
import contextlib
import io
import sys
import textwrap
from typing import List, Optional, Tuple, TextIO, Union, NoReturn

import icontract_hypothesis.pyicontract_hypothesis._general as _general
import icontract_hypothesis.pyicontract_hypothesis._ghostwrite as _ghostwrite
import icontract_hypothesis.pyicontract_hypothesis._test as _test


class Params:
    """Represent the parameters of the program."""

    def __init__(
        self, general: _general.Params, command: Union[_test.Params, _ghostwrite.Params]
    ) -> None:
        """Initialize with the given values."""
        self.general = general
        self.command = command


def _parse_args_to_params(
    args: argparse.Namespace,
) -> Tuple[Optional[Params], List[str]]:
    """
    Parse the parameters from the command-line arguments.

    Return parsed parameters, errors if any
    """
    errors = []  # type: List[str]

    general, general_errors = _general.parse_params(args=args)
    errors.extend(general_errors)

    command = None  # type: Optional[Union[_test.Params, _ghostwrite.Params]]
    if args.command == "test":
        test, command_errors = _test.parse_params(args=args)
        errors.extend(command_errors)

        command = test

    elif args.command == "ghostwrite":
        ghostwrite, command_errors = _ghostwrite.parse_params(args=args)
        errors.extend(command_errors)
        command = ghostwrite

    if errors:
        return None, errors

    assert general is not None
    assert command is not None

    return Params(general=general, command=command), []


def _make_argument_parser(
    prog: str = "pyicontract-hypothesis",
) -> argparse.ArgumentParser:
    """
    Create an instance of the argument parser to parse command-line arguments.

    :param prog: name of the program to be displayed in help
    """
    parser = argparse.ArgumentParser(prog=prog, description=__doc__)
    subparsers = parser.add_subparsers(help="Commands", dest="command")
    subparsers.required = True

    test_parser = subparsers.add_parser(
        "test",
        help="Test the functions automatically by inferring search strategies from preconditions",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    test_parser.add_argument(
        "-p", "--path", help="Path to the Python file to test", required=True
    )

    test_parser.add_argument(
        "--settings",
        help=textwrap.dedent(
            """\
            Specify settings for Hypothesis

            The settings are assigned by '='.
            The value of the setting needs to be encoded as JSON.

            Example: max_examples=500"""
        ),
        nargs="*",
    )

    test_parser.add_argument(
        "--inspect",
        help=textwrap.dedent(
            """\
            Only show the strategy and the settings

            No tests are executed."""
        ),
        action="store_true",
    )

    ghostwriter_parser = subparsers.add_parser(
        "ghostwrite",
        help="Ghostwrite the unit tests with inferred search strategies",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    input_group = ghostwriter_parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        "-m", "--module", help="Module to ghostwrite the unit tests for"
    )
    input_group.add_argument(
        "-p",
        "--path",
        help=textwrap.dedent(
            """\
            Path to the module to ghostwrite the unit tests for.

            If the file represents a module reachable through
            sys.path, use the qualified module name in
            the unit test.

            Otherwise, the module is represented as the stem
            of the path with all non-identifier characters
            replaced with an underscore ("_")."""
        ),
    )

    ghostwriter_parser.add_argument(
        "-o",
        "--output",
        help=textwrap.dedent(
            """\
            Path to the file where the output should be written.

            If '-', writes to STDOUT."""
        ),
        default="-",
    )

    ghostwriter_parser.add_argument(
        "--explicit",
        help=textwrap.dedent(
            """\
            Write the inferred strategies explicitly

            This is practical if you want to tune and
            refine the strategies and just want to use
            ghostwriting as a starting point.

            Mind that pyicontract-hypothesis does not
            automatically fix imports as this is
            usually project-specific. You have to fix imports
            manually after the ghostwriting."""
        ),
        action="store_true",
    )

    ghostwriter_parser.add_argument(
        "--bare",
        help=textwrap.dedent(
            """\
            Print only the body of the tests and omit header/footer
            (such as TestCase class or import statements).

            This is useful when you only want to inspect a single test or
            include a single test function in a custom test suite."""
        ),
        action="store_true",
    )

    for subparser in [test_parser, ghostwriter_parser]:
        subparser.add_argument(
            "-i",
            "--include",
            help=textwrap.dedent(
                """\
                Regular expressions, lines or line ranges of the functions to process

                If a line or line range overlaps the body of a function,
                the function is considered included.

                Example 1: ^do_something.*$
                Example 2: 3
                Example 3: 34-65"""
            ),
            required=False,
            nargs="*",
        )

        subparser.add_argument(
            "-e",
            "--exclude",
            help=textwrap.dedent(
                """\
                Regular expressions, lines or line ranges of the functions to exclude

                If a line or line range overlaps the body of a function,
                the function is considered excluded.

                Example 1: ^do_something.*$
                Example 2: 3
                Example 3: 34-65"""
            ),
            default=["^_.*$"],
            nargs="*",
        )

    return parser


def _parse_args(
    parser: argparse.ArgumentParser, argv: List[str]
) -> Tuple[Optional[argparse.Namespace], str, str]:
    """
    Parse the command-line arguments.

    Return (parsed args or None if failure, captured stdout, captured stderr).
    """
    pass  # for pydocstyle

    # From https://stackoverflow.com/questions/18160078
    @contextlib.contextmanager
    def captured_output():  # type: ignore
        new_out, new_err = io.StringIO(), io.StringIO()
        old_out, old_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout, sys.stderr
        finally:
            sys.stdout, sys.stderr = old_out, old_err

    with captured_output() as (out, err):
        try:
            parsed_args = parser.parse_args(argv)

            err.seek(0)
            out.seek(0)
            return parsed_args, out.read(), err.read()

        except SystemExit:
            err.seek(0)
            out.seek(0)
            return None, out.read(), err.read()


def assert_never(something: NoReturn) -> NoReturn:
    """Enforce exhaustive matching at mypy time."""
    assert False, "Unhandled type: {}".format(type(something).__name__)


def run(
    argv: List[str],
    stdout: TextIO,
    stderr: TextIO,
    prog: str = "pyicontract-hypothesis",
) -> int:
    """
    Execute the run routine.

    :param argv: list of command-line arguments
    :param stdout: stream to write the output to
    :param stderr: stream to write the error output to
    :param prog: name of the program to be displayed in help
    :return: exit code
    """
    parser = _make_argument_parser(prog=prog)
    args, out, err = _parse_args(parser=parser, argv=argv)
    if len(out) > 0:
        stdout.write(out)

    if len(err) > 0:
        stderr.write(err)

    if args is None:
        return 1

    params, errors = _parse_args_to_params(args=args)
    if errors:
        for error in errors:
            print(error, file=stderr)
            return 1

    assert params is not None

    if isinstance(params.command, _test.Params):
        errors = _test.test(
            general=params.general, command=params.command, stdout=stdout
        )
    elif isinstance(params.command, _ghostwrite.Params):
        code, errors = _ghostwrite.ghostwrite(
            general=params.general, command=params.command
        )
        if not errors:
            if params.command.output is None:
                stdout.write(code)
            else:
                params.command.output.write_text(code)
    else:
        assert_never(params.command)

    if errors:
        for error in errors:
            print(error, file=stderr)
            return 1

    return 0


def entry_point() -> int:
    """
    Wrap the entry_point routine wit default arguments.

    :return: exit code
    """
    return run(
        argv=sys.argv[1:],
        stdout=sys.stdout,
        stderr=sys.stderr,
        prog="pyicontract-hypothesis",
    )


if __name__ == "__main__":
    sys.exit(entry_point())



import datetime
import decimal
import fractions
import math
import sys
import unittest

import icontract

import icontract_hypothesis


class TestMatchingBounds(unittest.TestCase):
    def test_different_ops(self) -> None:
        hundred = 100

        @icontract.require(lambda x: x > 0)
        @icontract.require(lambda x: x >= 1)
        @icontract.require(lambda x: x < 100)
        @icontract.require(lambda x: x <= 90)
        @icontract.require(lambda y: 0 < y <= 100)
        @icontract.require(lambda y: 1 <= y < 90)
        @icontract.require(lambda z: 0 > z >= -math.sqrt(hundred))
        def some_func(x: int, y: int, z: int) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'x': integers(min_value=2, max_value=89),\n"
            " 'y': integers(min_value=2, max_value=89),\n"
            " 'z': integers(min_value=-9.0, max_value=-1)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_dates(self) -> None:
        SOME_DATE = datetime.date(2014, 3, 2)

        # The preconditions were picked s.t. to also test that we can recompute everything.
        @icontract.require(lambda a: a < SOME_DATE + datetime.timedelta(days=3))
        @icontract.require(lambda b: b < SOME_DATE + datetime.timedelta(days=2))
        @icontract.require(lambda c: c < max(SOME_DATE, datetime.date(2020, 1, 1)))
        @icontract.require(
            lambda d: d
            < (
                SOME_DATE
                if SOME_DATE > datetime.date(2020, 1, 1)
                else datetime.date(2020, 12, 5)
            )
        )
        def some_func(
            a: datetime.date, b: datetime.date, c: datetime.date, d: datetime.date
        ) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': dates(max_value=datetime.date(2014, 3, 4)),\n"
            " 'b': dates(max_value=datetime.date(2014, 3, 3)),\n"
            " 'c': dates(max_value=datetime.date(2019, 12, 31)),\n"
            " 'd': dates(max_value=datetime.date(2020, 12, 4))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_datetimes(self) -> None:
        SOME_DATETIME = datetime.datetime(2014, 3, 2, 10, 20, 30)

        @icontract.require(lambda a: a < SOME_DATETIME)
        def some_func(a: datetime.datetime) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': datetimes(max_value=datetime.datetime(2014, 3, 2, 10, 20, 29, 999999))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_times(self) -> None:
        SOME_TIME = datetime.time(1, 2, 3)

        @icontract.require(lambda a: a < SOME_TIME)
        def some_func(a: datetime.time) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'a': times(max_value=datetime.time(1, 2, 2, 999999))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_timedeltas(self) -> None:
        SOME_TIMEDELTA = datetime.timedelta(days=3)

        @icontract.require(lambda a: a < SOME_TIMEDELTA)
        def some_func(a: datetime.timedelta) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        if sys.version_info < (3, 7):
            self.assertEqual(
                "fixed_dictionaries("
                "{'a': timedeltas(max_value=datetime.timedelta(2, 86399, 999999))})",
                str(strategy),
            )
        else:
            self.assertEqual(
                "fixed_dictionaries({"
                "'a': timedeltas(max_value=datetime.timedelta("
                "days=2, seconds=86399, microseconds=999999))})",
                str(strategy),
            )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_fractions(self) -> None:
        SOME_FRACTION = fractions.Fraction(3, 2)

        @icontract.require(lambda a: a < SOME_FRACTION)
        def some_func(a: fractions.Fraction) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': fractions(max_value=Fraction(3, 2))"
            ".filter(lambda a: a < SOME_FRACTION)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_with_decimals(self) -> None:
        SOME_DECIMAL = decimal.Decimal(10)

        @icontract.require(lambda a: a < SOME_DECIMAL)
        @icontract.require(lambda a: not decimal.Decimal.is_nan(a))
        def some_func(a: decimal.Decimal) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries("
            "{'a': decimals(max_value=Decimal('10'))"
            ".filter(lambda a: not decimal.Decimal.is_nan(a))"
            ".filter(lambda a: a < SOME_DECIMAL)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)


if __name__ == "__main__":
    unittest.main()



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
import os
import sys

from setuptools import setup, find_packages

# pylint: disable=redefined-builtin

here = os.path.abspath(os.path.dirname(__file__))  # pylint: disable=invalid-name

with open(os.path.join(here, "README.rst"), encoding="utf-8") as fid:
    long_description = fid.read()  # pylint: disable=invalid-name

with open(os.path.join(here, "requirements.txt"), encoding="utf-8") as fid:
    install_requires = [line for line in fid.read().splitlines() if line.strip()]

setup(
    name="icontract-hypothesis",
    # Don't forget to update the version in __init__.py!
    version="1.1.4",
    description="Combine contracts and automatic testing.",
    long_description=long_description,
    url="https://github.com/mristin/icontract-hypothesis",
    author="Marko Ristin",
    author_email="marko@ristin.ch",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Hypothesis",
    ],
    license="License :: OSI Approved :: MIT License",
    keywords="design-by-contract contracts automatic testing property-based",
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    # fmt: off
    extras_require={
        "dev": [
            "black==20.8b1",
            "mypy==0.902",
            "pylint==2.3.1",
            "pydocstyle>=2.1.1,<3",
            "coverage>=4.5.1,<5",
            "docutils>=0.14,<1",
            "pygments>=2,<3",
            'dpcontracts==0.6.0',
            'hypothesis[dpcontracts]>=5,<6',
            'tabulate==0.8.7',
            'py-cpuinfo>=5,<6',
        ] + (['deal==4.4.0'] if sys.version_info >= (3, 8) else []),
    },
    # fmt: on
    py_modules=["icontract_hypothesis"],
    package_data={"icontract_hypothesis": ["py.typed"]},
    data_files=[(".", ["LICENSE", "README.rst", "requirements.txt"])],
    entry_points={
        "console_scripts": [
            "pyicontract-hypothesis = icontract_hypothesis.pyicontract_hypothesis.main:entry_point"
        ],
        "hypothesis": ["_ = icontract_hypothesis"],
    },
)



"""Link to the ``pyicontract_hypothesis`` so that you can execute it with ``python`` CLI."""

import sys

import icontract_hypothesis.pyicontract_hypothesis.main

if __name__ == "__main__":
    sys.exit(
        icontract_hypothesis.pyicontract_hypothesis.main.run(
            argv=sys.argv[1:],
            stdout=sys.stdout,
            stderr=sys.stderr,
            prog="icontract_hypothesis",
        )
    )



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=no-value-for-parameter


import re
import unittest

import icontract

import icontract_hypothesis


class TestMatchingRegex(unittest.TestCase):
    def test_re_match(self) -> None:
        @icontract.require(lambda s: re.match(r"^Start.*End$", s, flags=0))
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'s': from_regex(re.compile(r'^Start.*End$', re.UNICODE))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_re_renamed_match(self) -> None:
        import re as rerenamed

        @icontract.require(lambda s: rerenamed.match(r"^Start.*End$", s, flags=0))
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'s': from_regex(re.compile(r'^Start.*End$', re.UNICODE))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_multiple_re_match(self) -> None:
        @icontract.require(lambda s: re.match(r"^Start.*End$", s, flags=0))
        @icontract.require(lambda s: re.match(r"^.*something.*$", s, flags=0))
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            (
                "fixed_dictionaries("
                "{'s': from_regex(re.compile(r'^.*something.*$', re.UNICODE))"
                '.filter(lambda s: re.match(r"^Start.*End$", s, flags=0))})'
            ),
            str(strategy),
        )

        # We can not test this strategy as it is currently not matched.

    def test_compiled_re(self) -> None:
        START_END_RE = re.compile(r"^Start.*End$")

        @icontract.require(lambda s: START_END_RE.match(s))
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'s': from_regex(re.compile(r'^Start.*End$', re.UNICODE))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_compiled_re_with_logic(self) -> None:
        START_END_RE = re.compile(r"^Start.*End$")
        PREFIX_SUFFIX_RE = re.compile(r"^Prefix.*Suffix")

        SWITCH = False

        # This pre-condition also tests a bit more complicated logic so that we are sure that
        # the re-computation is executed successfully.
        @icontract.require(
            lambda s: (START_END_RE if SWITCH else PREFIX_SUFFIX_RE).match(s)
        )
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({'s': from_regex(re.compile(r'^Prefix.*Suffix', re.UNICODE))})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)

    def test_fullmatch(self) -> None:
        SOMETHING_RE = re.compile("something")

        # These assertions highlight the difference between ``match`` and ``fullmatch``.
        assert SOMETHING_RE.match("something else")
        assert not SOMETHING_RE.fullmatch("something else")
        assert SOMETHING_RE.fullmatch("something")

        @icontract.require(lambda s: SOMETHING_RE.fullmatch(s))
        def some_func(s: str) -> None:
            pass

        strategy = icontract_hypothesis.infer_strategy(some_func)
        self.assertEqual(
            "fixed_dictionaries({"
            "'s': from_regex(re.compile(r'something', re.UNICODE), fullmatch=True)})",
            str(strategy),
        )

        icontract_hypothesis.test_with_inferred_strategy(some_func)


if __name__ == "__main__":
    unittest.main()
